/*
 * This example demonstrates a basic level of abstraction of
 * a threading implementation.
 * This has been developed under Solaris, but should run on most
 * other POSIX compatible targets, too.
 * As multithreaded applications require reentrant system functions
 * you must pass -D_REENTRANT to metacc.
 * If your target uses the GNU compiler collection (gcc), then
 * you might additonally have to pass the following option to metacc:
 * -xgnu
 * 
 * For targets that do not support SCHED_FIFO properly (e.g. Linux)
 * you have to substitute the argument &task_a to pthread_create with 0.
 * 
 * On target that do not support the POSIX scheduling attribute "priority"
 * (e.g. Linux), you will have to comment out the following line:
 *	"param.sched_priority =  pri;"
 */


#include "internal.mh"

extern int usleep(unsigned int);
extern int printf(const char *, ...);

meta strg $Implementation;

meta struct task_attr
{
	zed priority;
	/* in theory, here could be additional attributes */
};

meta void $posix_thread(func, struct task_attr);
meta void $solaris_thread(func, struct task_attr);

meta void $static_task(func f, zed pri)
{
	struct task_attr a;
	a.priority = pri;
	switch($Implementation) {
	case "POSIX":
		$posix_thread(f,a);
		break;
	case "Solaris":
		$solaris_thread(f,a);
		break;
	default:
		$fatal("Unsupported implementation " + $Implementation + ". Possible value are POSIX, Solaris.");
	}
}

meta void $posix_thread(func f, struct task_attr a)
{
	symb task_id, task_a;
	symb entry_code, sync_code;
	zed pri = a.priority;
	
	create_task:= {
		struct sched_param param;
		bzero(&param,sizeof(struct sched_param));
		param.sched_priority =  pri;
		/* this is just an example, errorcodes must be checked for production use */
		pthread_attr_init(&task_a);
		pthread_attr_setschedpolicy(&task_a,PP_SCHED_FIFO);
		pthread_attr_setschedparam(&task_a, &param);
		if (pthread_create(&task_id,&task_a,f,0) != 0)
			perror("could not create thread");
	}
	join_task:=
		if (pthread_join(task_id,0) != 0)
			perror("could not join thread");
	$print("merging required headers...\n");
	$merge("pthread.mh");	/* as we use macro definitions from the file pthread.h during instantiation, we must do an abstraction */
	$merge("stdio.h");	/* for perror */
	$merge("strings.h");	/* for bzero */
	task_id = $define($module(),typeof($get_symbol($module(),"pthread_t")),(ident)(f) + "_tsk_id");
	task_a = $define($module(),(type) $get_symbol($module(),"pthread_attr_t"),(ident)(f) + "_tsk_attr");
	entry_code = $get_symbol($module(),(ident)"startup_tasks");
	$add_epilogue((func) entry_code, codeof(create_task));
	sync_code = $get_symbol($module(),(ident)"join_tasks");
	$add_epilogue((func) sync_code, codeof(join_task));
}


meta void $solaris_thread(func f, struct task_attr a)
{
	symb task_id, task_a;
	symb entry_code, sync_code;
	zed pri = a.priority;
	
	create_task:= {
		if (0 != thr_create(0,0,f,0,0,&task_id))
			perror("could not create thread");
		if (0 != thr_setprio(task_id, sched_get_priority_min(PP_SCHED_FIFO) + pri))
			perror("could not set thread priority");
	}
	join_task:=
		if (0 != thr_join(task_id,0,0))
			perror("could not join thread");
	$merge("sched.mh");
	$merge("thread.h");
	$merge("stdio.h");	/* for perror */
	task_id = $define($module(),typeof($get_symbol($module(),"thread_t")),(ident)(f) + "_tsk_id");
	entry_code = $get_symbol($module(),(ident)"startup_tasks");
	$add_stmt((stmt) entry_code, codeof(create_task));
	sync_code = $get_symbol($module(),(ident)"join_tasks");
	$add_stmt((stmt) sync_code, codeof(join_task));
}


void *heartbeat(void *ignored)
{
	while (1) {
		printf("heartbeat\n");
		usleep(300000);
		printf("heartbeat\n");
		usleep(700000);
	}
}


void startup_tasks(void)
{
	/* in this function all tasks are started */

}


void join_tasks(void)
{
	/* in this function all tasks are joined */

}


int main(void)
{
	startup_tasks();
	/* system is running */
	join_tasks();
}


meta void $main(strg arg[])
{
	if (arg[0])
		$Implementation = arg[0];
	else {
		$warn("implementation not set - using POSIX");
		$Implementation = "POSIX";
	}
	$static_task(heartbeat,0);
}

/* vim:filetype=c
 * vim:tw=0
 */
