/*
 * Copyright 2005-2007, Thomas Maier-Komor
 *
 * This is a full-featured metaprogram, implementing worst-case-execution-time
 * (WCET) measurement instrumentation for arbitrary C source code.
 * This implementation uses clock which is very unprecises. Better would be
 * Solaris' "gethrtime" to get the timing information. If you want to use it
 * for a program that is targeted towards another operating system, you must
 * adjust the measurement function accordingly.
 *
 * The metaprogram supports the adjustion of return statements to increase 
 * measurement accuracy. The included C source code will be instrumented by the
 * metaprogram.
 *
 * You are allowed to use this metaprogram to instrument and analyse any code,
 * provided that you stick to the license of the MetaC Compiler. Additionally,
 * if you publish any results obtained by use of this metaprogram, you must
 * reference MetaC and this metaprogram in your publication.
 */


#include "internal.mh"
#include "vc_hacks.h"
#include <time.h>

#include "gzip-1.2.4a/deflate.c"

meta symb $mt, $mdata, $m_at;

/* This metafunction creates a temporary variable, to which
 * the result of the return expression is assigned. The return
 * is adjusted to return the temporary variable. Like this
 * it is possible to measure the execution time of a function more
 * accurately.
 * Be aware that on SPARC based systems this will disable 
 * tail-call-optimization. So you might decide to enhance this
 * metafunction to detect these situations. This enhancement is left
 * to you as a suggested exercise...
 */
meta void $return_from_var(func f)
{
	stmt rets[], *r, *er, ns;
	expr re,ne;
	symb tmp;
	type rt;

	rt = $get_return_type(f);
	er = $get_returns(f,rets);
	if (rt == typeof(void))
		return;
	tmp = $define((scope)f,rt,(ident)"return_temporary");
	ne = (expr)tmp;
	r = rets;
	$print("adjusting "+(strg)(er-r)+" return statements in function "+(strg) f+"\n");
	while (*r) {
		re = $get_return_expr(*r);
		assign_tmp:= tmp = re;
		ns = $insert_before(codeof assign_tmp,*r);
		$print("inserted "+(strg)ns);
		$set_return_expr(*r,ne);

//		ret_tmp := return tmp;
//		ns = $replace_stmt(*r,codeof assign_tmp);
//		$add_stmt($parent(ns),codeof ret_tmp);

		++r;
	}
}


/* This metafunction instruments the function f. Parameter i is the first
 * unique identifier number available.
 */
meta zed $instrument_function(func f, zed $i)
{
	stmt branches[], *branch, body, parents[], *xp, bc, p, pred;
	strg op;
	measure:=
		if ($m_at < (sizeof($mdata)/sizeof(measure_data_t))) {
			$mdata[$m_at].mp = $i, $mdata[$m_at].ts = clock();
			++$m_at;
		}
  
	if (f == 0)
		return $i;
	$get_all_branches(f,branches);
	branch = branches;
	while (*branch) {
		bc = $branch_creator(*branch);
		xp = parents;
		while (*xp && (*xp != bc))
			++xp;
		if (!*xp) {
			*xp = bc;
			pred = $predecessor(bc);
			p = $parent(bc);
			if ((pred && !$is_branch(pred)) || ((pred == 0) && !$is_branch(p) && $parent(p))) {
				//op = generate_stmt($i);
				$before_branch(codeof measure,*branch);
				++$i;
			}
		}
		//op = generate_stmt($i);
		*branch = $end(codeof(measure),*branch);
		$info("added measurement point " + (strg)($i) + " to function " + (strg) f);
		++$i;
		++branch;
	}
	body = $get_body(f);
	//op = generate_stmt($i);
	++$i;
	$begin(codeof(measure), body);
	//op = generate_stmt($i);
	++$i;
	$end(codeof(measure), body);
	return $i;
}


/* This metafunction searches for expressions which might generate machine code
 * with branches in its control flow and warns the user that these instances
 * will not be instrumented.
 */
meta void $verify_expressions(func f)
{
	stmt stmts[], *st;

	st = stmts;
	$info("verify_expressions: verifying function " + $name((symb)f));
	$get_all_stmts(st,f);
	while (*st) {
		expr e;
		stmt s = *st;
		++st;
		e = $get_expr(s);
		if (!e)
			continue;
		if ($has_cfbranch(e)) {
			$warn("statement in line " + (strg) $line_of(s) + " has an expression with branches in control flow");
			continue;
		}
		if ($is_for(s)) {
			e = $get_for_init(s);
			if (e && $has_cfbranch(e)) {
				$warn("statement in line " + (strg) $line_of(s) + " has an expression with branches in control flow");
				continue;
			}
			e = $get_for_iterator(s);
			if (e && $has_cfbranch(e)) {
				$warn("statement in line " + (strg) $line_of(s) + " has an expression with branches in control flow");
				continue;
			}
		}
	}
}


meta void $main(strg arg[])
{
	zed i = 0, s = 0;
	func all_func[], *f = all_func;
	symb pr, m;
       	func mf;

	s = (zed)arg[0];
	if (s < 1) {
		$warn("Please take a look at this code before trying to compile it!\n");
		s = 4096;
	}
	/* add headers defining functions and types we will use */
	$merge("stdio.h");
	$merge("stdlib.h");

	/* define types and variables that we need */
	$mt = $define_type($module(), typeof(struct {long long ts; unsigned mp; }), (ident)"measure_data_t");
	$mdata = $define($module(), typeof(typedef $mt[]), (ident)"mdata");
	$resize_array($mdata,s);
	$m_at = $define($module(), typeof(unsigned), (ident)"m_at",0);
	
	$get_all_functions(all_func);
	while (*f) {
		$verify_expressions(*f);
		$return_from_var(*f);
		i = $instrument_function(*f,i);
		++f;
	}
	$print("instanciated " + (strg)i + " measure-points\n");

	print_result := {
				int j;
				for (j = 0; j < sizeof(mdata)/sizeof(measure_data_t); ++j) {
					printf("%10lld : %8ld\n", $mdata[j].ts, $mdata[j].mp);
				}
			}
	pr = $define($module(),typeof(void (void)),(ident)"print_measure_results");
	$implement(pr, codeof print_result);

	m = $get_symbol($module(),(ident)"main");
	mf = (func)m;
	if (mf == 0) {
		$warn("Function main isn't visible, defining a trunk. Please do the merge yourself.");
		m = $define($module(),typeof(int(void)),(ident)"main");
		empty := {}
		mf = $implement(m,codeof empty);
	}
	print_results := print_measure_results();
	$add_epilogue(mf,codeof print_results);
}

/* vim:tw=0
 * vim:filetype=c
 */
